use crate::prelude::*;
use crate::samplers::sampler::*;
use std::f32::consts::PI;

/**
 * Samples a segment of a circle.
 * Angles are as on a compass, starting with 0 at North and increasing clockwise.
 * North is Vector2(0.0, -1.0), East is Vector2(1.0, 0.0)
 */
pub struct CircularAlignmentSampler {
    lower_end_rad: f32,
    upper_end_rad: f32,
    center:        Vector2,
    radius_meters: f32,
    clockwise:     bool,
}

impl CircularAlignmentSampler {
    pub fn new(mut lower_end_rad: f32, mut upper_end_rad: f32, center: Vector2, radius_meters: f32, clockwise: bool) -> Self {
        if clockwise && upper_end_rad < lower_end_rad {
            upper_end_rad += 2.0 * PI;
        } else if !clockwise && upper_end_rad > lower_end_rad {
            lower_end_rad += 2.0 * PI;
        }
        Self {
            lower_end_rad,
            upper_end_rad,
            center,
            radius_meters,
            clockwise,
        }
    }
    
    fn angle_from_normalized(&self, t: f32) -> f32 {
        self.lower_end_rad * (1.0 - t) + self.upper_end_rad * t
    }
}

impl AlignmentSampler for CircularAlignmentSampler {
    fn sample_normalized(&self, t: f32) -> Vector2 {
        let t_rad = self.angle_from_normalized(t);
        return Vector2::new(t_rad.sin() * self.radius_meters, -t_rad.cos() * self.radius_meters) + self.center;
    }
    
    fn tangent_normalized(&self, t: f32) -> Vector2 {
        let t_rad = self.angle_from_normalized(t);
        if self.clockwise {
            return Vector2::new(t_rad.cos(), t_rad.sin());
        } else {
            return Vector2::new(-t_rad.cos(), -t_rad.sin());
        }
    }
    
    fn sample_offset(&self, offset: f32) -> Vector2 {
        self.sample_normalized(offset / self.length())
    }
    
    fn tangent_offset(&self, offset: f32) -> Vector2 {
        self.tangent_normalized(offset / self.length())
    }
    
    fn length(&self) -> f32 {
        (self.upper_end_rad - self.lower_end_rad).abs() * self.radius_meters
    }
}

pub struct CircularElevationSampler {
    lower_end: f32,
    upper_end: f32,
    center_meters: f32,
    radius_meters: f32,
    top_down_length: f32,
    peak:   bool, // as opposed to valley
}

impl CircularElevationSampler {
    pub fn new(lower_end: f32,
           upper_end: f32,
           center_meters: f32,
           radius_meters: f32,
           peak:   bool,) -> Self {
        let top_down_length = (upper_end - lower_end) * radius_meters;
        Self {
            lower_end,
            upper_end,
            center_meters,
            radius_meters,
            top_down_length,
            peak,
        }
    }
}

fn circle_function(x: f32) -> f32 {
    let x = x.clamp(-1.0, 1.0);
    (1.0 - x * x).sqrt()
}

impl ElevationSampler for CircularElevationSampler {
    fn sample_normalized(&self, top_down_t: f32) -> f32 {
        let actual_t = self.lower_end * (1.0 - top_down_t) + self.upper_end * top_down_t;
        if self.peak {
            return circle_function(actual_t) * self.radius_meters + self.center_meters;
        } else {
            return -circle_function(actual_t) * self.radius_meters + self.center_meters;
        }
    }
    
    fn grade_normalized(&self, top_down_t: f32) -> f32 {
        let actual_t = self.lower_end * (1.0 - top_down_t) + self.upper_end * top_down_t;
        if self.peak {
            return -actual_t / circle_function(actual_t);
        } else {
            return actual_t / circle_function(actual_t);
        }
    }
    
    fn sample_offset(&self, offset: f32) -> f32 {
        self.sample_normalized(offset / self.top_down_length)
    }
    
    fn grade_offset(&self, offset: f32) -> f32 {
        self.grade_normalized(offset / self.top_down_length)
    }
    
    fn length(&self) -> f32 {
        (self.upper_end.asin() - self.lower_end.asin()) * self.radius_meters
    }
    
    fn to_top_down_normalized(&self, actual_t: f32) -> f32 {
        let actual_t = self.lower_end.asin() * (1.0 - actual_t) + self.upper_end.asin() * actual_t;
        (actual_t.sin() - self.lower_end) / (self.upper_end - self.lower_end)
    }
    
    fn to_top_down_offset(&self, actual_offset: f32) -> f32 {
        self.to_top_down_normalized(actual_offset / self.length()) * self.top_down_length
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    
    fn compare_vectors(vec1: Vector2, vec2: Vector2, message: &str) {
        let epsilon = 1e-5; // This is really bad tolerance, but f32 sucks so I have no choice.
        assert!((vec1.x - vec2.x).abs() <= epsilon && (vec1.y - vec2.y).abs() <= epsilon,
                "{message} Expected {vec2}, received {vec1}.")
    }
    
    fn compare_floats(float1: f32, float2: f32, message: &str) {
        let epsilon = 1e-5;
        assert!((float1 - float2).abs() <= epsilon,
                "{message} Expected {float2}, received {float1}.")
    }
    
    #[test]
    fn test_circular_alignment_sample_normalized() {
        let test_alignment_1 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_1.sample_normalized(0.0), Vector2::new(1.0, -1.0),
                   "test_alignment_1 begins in the wrong location.");
        compare_vectors(test_alignment_1.sample_normalized(1.0), Vector2::new(1.0, 3.0),
                   "test_alignment_1 ends in the wrong location.");
        compare_vectors(test_alignment_1.sample_normalized(0.5), Vector2::new(3.0, 1.0),
                   "test_alignment_1 passes through the wrong location.");
        compare_vectors(test_alignment_1.sample_normalized(-1.0), Vector2::new(1.0, 3.0),
                   "test_alignment_1 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_1.sample_normalized(2.0), Vector2::new(1.0, -1.0),
                   "test_alignment_1 double offset ends in the wrong location.");
        let test_alignment_2 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_2.sample_normalized(0.0), Vector2::new(-1.0, 1.0),
                   "test_alignment_2 begins in the wrong location.");
        compare_vectors(test_alignment_2.sample_normalized(1.0), Vector2::new(3.0, 1.0),
                   "test_alignment_2 ends in the wrong location.");
        compare_vectors(test_alignment_2.sample_normalized(0.5), Vector2::new(1.0, -1.0),
                   "test_alignment_2 passes through the wrong location.");
        compare_vectors(test_alignment_2.sample_normalized(-1.0), Vector2::new(3.0, 1.0),
                   "test_alignment_2 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_2.sample_normalized(2.0), Vector2::new(-1.0, 1.0),
                   "test_alignment_2 double offset ends in the wrong location.");
        let test_alignment_3 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_3.sample_normalized(0.0), Vector2::new(1.0, -1.0),
                   "test_alignment_3 begins in the wrong location.");
        compare_vectors(test_alignment_3.sample_normalized(1.0), Vector2::new(1.0, 3.0),
                   "test_alignment_3 ends in the wrong location.");
        compare_vectors(test_alignment_3.sample_normalized(0.5), Vector2::new(-1.0, 1.0),
                   "test_alignment_3 passes through the wrong location.");
        compare_vectors(test_alignment_3.sample_normalized(-1.0), Vector2::new(1.0, 3.0),
                   "test_alignment_3 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_3.sample_normalized(2.0), Vector2::new(1.0, -1.0),
                   "test_alignment_3 double offset ends in the wrong location.");
        let test_alignment_4 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_4.sample_normalized(0.0), Vector2::new(-1.0, 1.0),
                   "test_alignment_4 begins in the wrong location.");
        compare_vectors(test_alignment_4.sample_normalized(1.0), Vector2::new(3.0, 1.0),
                   "test_alignment_4 ends in the wrong location.");
        compare_vectors(test_alignment_4.sample_normalized(0.5), Vector2::new(1.0, 3.0),
                   "test_alignment_4 passes through the wrong location.");
        compare_vectors(test_alignment_4.sample_normalized(-1.0), Vector2::new(3.0, 1.0),
                   "test_alignment_4 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_4.sample_normalized(2.0), Vector2::new(-1.0, 1.0),
                   "test_alignment_4 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_circular_alignment_tangent_normalized() {
        let test_alignment_1 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_1.tangent_normalized(0.0), Vector2::new(1.0, 0.0),
                   "test_alignment_1 begins in the wrong direction.");
        compare_vectors(test_alignment_1.tangent_normalized(1.0), Vector2::new(-1.0, 0.0),
                   "test_alignment_1 ends in the wrong direction.");
        compare_vectors(test_alignment_1.tangent_normalized(0.5), Vector2::new(0.0, 1.0),
                   "test_alignment_1 passes through the wrong direction.");
        compare_vectors(test_alignment_1.tangent_normalized(-1.0), Vector2::new(-1.0, 0.0),
                   "test_alignment_1 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_1.tangent_normalized(2.0), Vector2::new(1.0, 0.0),
                   "test_alignment_1 double offset ends in the wrong direction.");
        let test_alignment_2 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_2.tangent_normalized(0.0), Vector2::new(0.0, -1.0),
                   "test_alignment_2 begins in the wrong direction.");
        compare_vectors(test_alignment_2.tangent_normalized(1.0), Vector2::new(0.0, 1.0),
                   "test_alignment_2 ends in the wrong direction.");
        compare_vectors(test_alignment_2.tangent_normalized(0.5), Vector2::new(1.0, 0.0),
                   "test_alignment_2 passes through the wrong direction.");
        compare_vectors(test_alignment_2.tangent_normalized(-1.0), Vector2::new(0.0, 1.0),
                   "test_alignment_2 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_2.tangent_normalized(2.0), Vector2::new(0.0, -1.0),
                   "test_alignment_2 double offset ends in the wrong direction.");
        let test_alignment_3 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_3.tangent_normalized(0.0), Vector2::new(-1.0, 0.0),
                   "test_alignment_3 begins in the wrong direction.");
        compare_vectors(test_alignment_3.tangent_normalized(1.0), Vector2::new(1.0, 0.0),
                   "test_alignment_3 ends in the wrong direction.");
        compare_vectors(test_alignment_3.tangent_normalized(0.5), Vector2::new(0.0, 1.0),
                   "test_alignment_3 passes through the wrong direction.");
        compare_vectors(test_alignment_3.tangent_normalized(-1.0), Vector2::new(1.0, 0.),
                   "test_alignment_3 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_3.tangent_normalized(2.0), Vector2::new(-1.0, 0.0),
                   "test_alignment_3 double offset ends in the wrong direction.");
        let test_alignment_4 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_4.tangent_normalized(0.0), Vector2::new(0.0, 1.0),
                   "test_alignment_4 begins in the wrong direction.");
        compare_vectors(test_alignment_4.tangent_normalized(1.0), Vector2::new(0.0, -1.0),
                   "test_alignment_4 ends in the wrong direction.");
        compare_vectors(test_alignment_4.tangent_normalized(0.5), Vector2::new(1.0, 0.0),
                   "test_alignment_4 passes through the wrong direction.");
        compare_vectors(test_alignment_4.tangent_normalized(-1.0), Vector2::new(0.0, -1.0),
                   "test_alignment_4 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_4.tangent_normalized(2.0), Vector2::new(0.0, 1.0),
                   "test_alignment_4 double offset ends in the wrong direction.");
    }
    
    #[test]
    fn test_circular_alignment_sample_offset() {
        let test_alignment_1 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_1.sample_offset(0.0), Vector2::new(1.0, -1.0),
                   "test_alignment_1 begins in the wrong location.");
        compare_vectors(test_alignment_1.sample_offset(2.0 * PI), Vector2::new(1.0, 3.0),
                   "test_alignment_1 ends in the wrong location.");
        compare_vectors(test_alignment_1.sample_offset(PI), Vector2::new(3.0, 1.0),
                   "test_alignment_1 passes through the wrong location.");
        compare_vectors(test_alignment_1.sample_offset(-2.0 * PI), Vector2::new(1.0, 3.0),
                   "test_alignment_1 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_1.sample_offset(4.0 * PI), Vector2::new(1.0, -1.0),
                   "test_alignment_1 double offset ends in the wrong location.");
        let test_alignment_2 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_2.sample_offset(0.0), Vector2::new(-1.0, 1.0),
                   "test_alignment_2 begins in the wrong location.");
        compare_vectors(test_alignment_2.sample_offset(2.0 * PI), Vector2::new(3.0, 1.0),
                   "test_alignment_2 ends in the wrong location.");
        compare_vectors(test_alignment_2.sample_offset(PI), Vector2::new(1.0, -1.0),
                   "test_alignment_2 passes through the wrong location.");
        compare_vectors(test_alignment_2.sample_offset(-2.0 * PI), Vector2::new(3.0, 1.0),
                   "test_alignment_2 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_2.sample_offset(4.0 * PI), Vector2::new(-1.0, 1.0),
                   "test_alignment_2 double offset ends in the wrong location.");
        let test_alignment_3 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_3.sample_offset(0.0), Vector2::new(1.0, -1.0),
                   "test_alignment_3 begins in the wrong location.");
        compare_vectors(test_alignment_3.sample_offset(2.0 * PI), Vector2::new(1.0, 3.0),
                   "test_alignment_3 ends in the wrong location.");
        compare_vectors(test_alignment_3.sample_offset(PI), Vector2::new(-1.0, 1.0),
                   "test_alignment_3 passes through the wrong location.");
        compare_vectors(test_alignment_3.sample_offset(-2.0 * PI), Vector2::new(1.0, 3.0),
                   "test_alignment_3 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_3.sample_offset(4.0 * PI), Vector2::new(1.0, -1.0),
                   "test_alignment_3 double offset ends in the wrong location.");
        let test_alignment_4 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_4.sample_offset(0.0), Vector2::new(-1.0, 1.0),
                   "test_alignment_4 begins in the wrong location.");
        compare_vectors(test_alignment_4.sample_offset(2.0 * PI), Vector2::new(3.0, 1.0),
                   "test_alignment_4 ends in the wrong location.");
        compare_vectors(test_alignment_4.sample_offset(PI), Vector2::new(1.0, 3.0),
                   "test_alignment_4 passes through the wrong location.");
        compare_vectors(test_alignment_4.sample_offset(-2.0 * PI), Vector2::new(3.0, 1.0),
                   "test_alignment_4 negative offset ends in the wrong location.");
        compare_vectors(test_alignment_4.sample_offset(4.0 * PI), Vector2::new(-1.0, 1.0),
                   "test_alignment_4 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_circular_alignment_tangent_offset() {
        let test_alignment_1 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_1.tangent_offset(0.0), Vector2::new(1.0, 0.0),
                   "test_alignment_1 begins in the wrong direction.");
        compare_vectors(test_alignment_1.tangent_offset(2.0 * PI), Vector2::new(-1.0, 0.0),
                   "test_alignment_1 ends in the wrong direction.");
        compare_vectors(test_alignment_1.tangent_offset(PI), Vector2::new(0.0, 1.0),
                   "test_alignment_1 passes through the wrong direction.");
        compare_vectors(test_alignment_1.tangent_offset(-2.0 * PI), Vector2::new(-1.0, 0.0),
                   "test_alignment_1 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_1.tangent_offset(4.0 * PI), Vector2::new(1.0, 0.0),
                   "test_alignment_1 double offset ends in the wrong direction.");
        let test_alignment_2 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, true);
        compare_vectors(test_alignment_2.tangent_offset(0.0), Vector2::new(0.0, -1.0),
                   "test_alignment_2 begins in the wrong direction.");
        compare_vectors(test_alignment_2.tangent_offset(2.0 * PI), Vector2::new(0.0, 1.0),
                   "test_alignment_2 ends in the wrong direction.");
        compare_vectors(test_alignment_2.tangent_offset(PI), Vector2::new(1.0, 0.0),
                   "test_alignment_2 passes through the wrong direction.");
        compare_vectors(test_alignment_2.tangent_offset(-2.0 * PI), Vector2::new(0.0, 1.0),
                   "test_alignment_2 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_2.tangent_offset(4.0 * PI), Vector2::new(0.0, -1.0),
                   "test_alignment_2 double offset ends in the wrong direction.");
        let test_alignment_3 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_3.tangent_offset(0.0), Vector2::new(-1.0, 0.0),
                   "test_alignment_3 begins in the wrong direction.");
        compare_vectors(test_alignment_3.tangent_offset(2.0 * PI), Vector2::new(1.0, 0.0),
                   "test_alignment_3 ends in the wrong direction.");
        compare_vectors(test_alignment_3.tangent_offset(PI), Vector2::new(0.0, 1.0),
                   "test_alignment_3 passes through the wrong direction.");
        compare_vectors(test_alignment_3.tangent_offset(-2.0 * PI), Vector2::new(1.0, 0.),
                   "test_alignment_3 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_3.tangent_offset(4.0 * PI), Vector2::new(-1.0, 0.0),
                   "test_alignment_3 double offset ends in the wrong direction.");
        let test_alignment_4 = CircularAlignmentSampler::new(3.0 * PI / 2.0, PI / 2.0, Vector2::new(1.0, 1.0), 2.0, false);
        compare_vectors(test_alignment_4.tangent_offset(0.0), Vector2::new(0.0, 1.0),
                   "test_alignment_4 begins in the wrong direction.");
        compare_vectors(test_alignment_4.tangent_offset(2.0 * PI), Vector2::new(0.0, -1.0),
                   "test_alignment_4 ends in the wrong direction.");
        compare_vectors(test_alignment_4.tangent_offset(PI), Vector2::new(1.0, 0.0),
                   "test_alignment_4 passes through the wrong direction.");
        compare_vectors(test_alignment_4.tangent_offset(-2.0 * PI), Vector2::new(0.0, -1.0),
                   "test_alignment_4 negative offset ends in the wrong direction.");
        compare_vectors(test_alignment_4.tangent_offset(4.0 * PI), Vector2::new(0.0, 1.0),
                   "test_alignment_4 double offset ends in the wrong direction.");
    }
    
    #[test]
    fn test_circular_alignment_length() {
        let test_alignment_1 = CircularAlignmentSampler::new(0.0, PI, Vector2::new(1.0, 1.0), 2.0, true);
        compare_floats(test_alignment_1.length(), PI * 2.0, "test_alignment_1 had an incorrect length.");
    }
    
    #[test]
    fn test_circular_elevation_sample_normalized() {
        let test_elevation_1 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, true);
        compare_floats(test_elevation_1.sample_normalized(0.0), 1.0 + 3.0_f32.sqrt(),
                   "test_elevation_1 begins in the wrong location.");
        compare_floats(test_elevation_1.sample_normalized(1.0), 1.0 + 3.0_f32.sqrt(),
                   "test_elevation_1 ends in the wrong location.");
        compare_floats(test_elevation_1.sample_normalized(0.5), 3.0,
                   "test_elevation_1 passes through the wrong location.");
        let test_elevation_2 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, false);
        compare_floats(test_elevation_2.sample_normalized(0.0), 1.0 - 3.0_f32.sqrt(),
                   "test_elevation_2 begins in the wrong location.");
        compare_floats(test_elevation_2.sample_normalized(1.0), 1.0 - 3.0_f32.sqrt(),
                   "test_elevation_2 ends in the wrong location.");
        compare_floats(test_elevation_2.sample_normalized(0.5), -1.0,
                   "test_elevation_2 passes through the wrong location.");
    }
    
    #[test]
    fn test_circular_elevation_grade_normalized() {
        let test_elevation_1 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, true);
        compare_floats(test_elevation_1.grade_normalized(0.0), 1.0 / 3.0_f32.sqrt(),
                   "test_elevation_1 begins in the wrong grade.");
        compare_floats(test_elevation_1.grade_normalized(1.0), -1.0 / 3.0_f32.sqrt(),
                   "test_elevation_1 ends in the wrong grade.");
        compare_floats(test_elevation_1.grade_normalized(0.5), 0.0,
                   "test_elevation_1 passes through the wrong grade.");
        let test_elevation_2 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, false);
        compare_floats(test_elevation_2.grade_normalized(0.0), -1.0 / 3.0_f32.sqrt(),
                   "test_elevation_2 begins in the wrong grade.");
        compare_floats(test_elevation_2.grade_normalized(1.0), 1.0 / 3.0_f32.sqrt(),
                   "test_elevation_2 ends in the wrong grade.");
        compare_floats(test_elevation_2.grade_normalized(0.5), 0.0,
                   "test_elevation_2 passes through the wrong grade.");
    }
    
    #[test]
    fn test_circular_elevation_sample_offset() {
        let test_elevation_1 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, true);
        compare_floats(test_elevation_1.sample_offset(0.0), 1.0 + 3.0_f32.sqrt(),
                   "test_elevation_1 begins in the wrong location.");
        compare_floats(test_elevation_1.sample_offset(2.0), 1.0 + 3.0_f32.sqrt(),
                   "test_elevation_1 ends in the wrong location.");
        compare_floats(test_elevation_1.sample_offset(1.0), 3.0,
                   "test_elevation_1 passes through the wrong location.");
        let test_elevation_2 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, false);
        compare_floats(test_elevation_2.sample_offset(0.0), 1.0 - 3.0_f32.sqrt(),
                   "test_elevation_2 begins in the wrong location.");
        compare_floats(test_elevation_2.sample_offset(2.0), 1.0 - 3.0_f32.sqrt(),
                   "test_elevation_2 ends in the wrong location.");
        compare_floats(test_elevation_2.sample_offset(1.0), -1.0,
                   "test_elevation_2 passes through the wrong location.");
    }
    
    #[test]
    fn test_circular_elevation_grade_offset() {
        let test_elevation_1 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, true);
        compare_floats(test_elevation_1.grade_offset(0.0), 1.0 / 3.0_f32.sqrt(),
                   "test_elevation_1 begins in the wrong grade.");
        compare_floats(test_elevation_1.grade_offset(2.0), -1.0 / 3.0_f32.sqrt(),
                   "test_elevation_1 ends in the wrong grade.");
        compare_floats(test_elevation_1.grade_offset(1.0), 0.0,
                   "test_elevation_1 passes through the wrong grade.");
        let test_elevation_2 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, false);
        compare_floats(test_elevation_2.grade_offset(0.0), -1.0 / 3.0_f32.sqrt(),
                   "test_elevation_2 begins in the wrong grade.");
        compare_floats(test_elevation_2.grade_offset(2.0), 1.0 / 3.0_f32.sqrt(),
                   "test_elevation_2 ends in the wrong grade.");
        compare_floats(test_elevation_2.grade_offset(1.0), 0.0,
                   "test_elevation_2 passes through the wrong grade.");
    }
    
    #[test]
    fn test_circular_elevation_length() {
        let test_elevation_1 = CircularElevationSampler::new(-0.5, 0.5, 1.0, 2.0, true);
        compare_floats(test_elevation_1.length(), 2.0 * PI / 3.0, "test_elevation_1 had an incorrect length.");
    }
    
    #[test]
    fn test_circular_elevation_to_top_down_normalized() {
        let test_elevation_1 = CircularElevationSampler::new(-1.0, 1.0, 1.0, 2.0, true);
        compare_floats(test_elevation_1.to_top_down_normalized(0.0), 0.0,
                   "test_elevation_1 begins in the wrong location.");
        compare_floats(test_elevation_1.to_top_down_normalized(1.0), 1.0,
                   "test_elevation_1 ends in the wrong location.");
        compare_floats(test_elevation_1.to_top_down_normalized(0.5), 0.5,
                   "test_elevation_1 passes through the wrong location.");
        compare_floats(test_elevation_1.to_top_down_normalized(0.25), (1.0 - (2.0_f32.sqrt() / 2.0)) / 2.0,
                   "test_elevation_1 quarter offset ends in the wrong location.");
        compare_floats(test_elevation_1.to_top_down_normalized(0.75), ((2.0_f32.sqrt() / 2.0) + 1.0) / 2.0,
                   "test_elevation_1 three quarter offset ends in the wrong location.");
    }
    
    #[test]
    fn test_circular_elevation_to_top_down_offset() {
        let test_elevation_1 = CircularElevationSampler::new(-1.0, 1.0, 1.0, 2.0, true);
        compare_floats(test_elevation_1.to_top_down_offset(0.0), 0.0,
                   "test_elevation_1 begins in the wrong location.");
        compare_floats(test_elevation_1.to_top_down_offset(2.0 * PI), 4.0,
                   "test_elevation_1 ends in the wrong location.");
        compare_floats(test_elevation_1.to_top_down_offset(PI), 2.0,
                   "test_elevation_1 passes through the wrong location.");
        compare_floats(test_elevation_1.to_top_down_offset(PI / 2.0), (1.0 - (2.0_f32.sqrt() / 2.0)) * 2.0,
                   "test_elevation_1 quarter offset ends in the wrong location.");
        compare_floats(test_elevation_1.to_top_down_offset(3.0 * PI / 2.0), ((2.0_f32.sqrt() / 2.0) + 1.0) * 2.0,
                   "test_elevation_1 three quarter offset ends in the wrong location.");
    }
}
