use crate::prelude::*;

/**
 * Implement this for 2D top-down point samplers.
 * These samplers will be used for a segment's top-down shape.
 */
pub trait AlignmentSampler {
    
    /**
     * Input distance from lower end in percent, get the top-down coordinates.
     */
    fn sample_normalized(&self, t: f32) -> Vector2;
    
    /**
     * Input distance from lower end in percent, get the normalized direction of the curve at that point.
     */
    fn tangent_normalized(&self, t: f32) -> Vector2;
    
    /**
     * Input distance from lower end in meters, get the top-down coordinates.
     */
    fn sample_offset(&self, offset: f32) -> Vector2;
    
    /**
     * Input distance from lower end in meters, get the normalized direction of the curve at that point.
     */
    fn tangent_offset(&self, offset: f32) -> Vector2;
    
    /**
     * Get the length as seen from above. This does not account for elevation changes.
     */
    fn length(&self) -> f32;
}

/**
 * Implement this for 1D elevation point samplers.
 * These samplers will be used for a segment's elevation or superelevation.
 * Structs that implement this should keep track of their top-down length.
 */
pub trait ElevationSampler {
    
    /**
     * Input normalized distance from lower end, get elevation at that point.
     */
    fn sample_normalized(&self, top_down_t: f32) -> f32;
    
    /**
     * Input normalized distance from lower end, get the derivative of elevation at that point.
     * This is equivalent to the track's grade at this point.
     */
    fn grade_normalized(&self, top_down_t: f32) -> f32;
    
    /**
     * Input actual distance in meters from lower end, get elevation at that point.
     */
    fn sample_offset(&self, top_down_offset: f32) -> f32;
    
    /**
     * Input actual distance in meters from lower end, get the derivative of elevation at that point.
     * This is equivalent to the track's grade at this point.
     */
    fn grade_offset(&self, top_down_offset: f32) -> f32;
    
    /**
     * Get the actual length accounting for elevation.
     */
    fn length(&self) -> f32;
    
    /**
     * Given an offset relative to the total length of the segment, return the associated distance.
     */
    fn to_top_down_normalized(&self, actual_t: f32) -> f32;
    
    /**
     * Given an offset relative to the total length of the segment, return the associated distance.
     */
    fn to_top_down_offset(&self, actual_offset: f32) -> f32;
}
