use crate::prelude::*;
use crate::samplers::sampler::*;

pub struct TangentAlignmentSampler {
    lower_end: Vector2,
    upper_end: Vector2,
}

impl TangentAlignmentSampler {
    pub fn new(lower_end: Vector2, upper_end: Vector2) -> Self {
        Self {lower_end, upper_end}
    }
}

impl AlignmentSampler for TangentAlignmentSampler {
    fn sample_normalized(&self, t: f32) -> Vector2 {
        self.lower_end.lerp(self.upper_end, t)
    }
    
    fn tangent_normalized(&self, _t: f32) -> Vector2 {
        (self.upper_end - self.lower_end).normalized()
    }
    
    fn sample_offset(&self, offset: f32) -> Vector2 {
        self.sample_normalized(offset / self.length())
    }
    
    fn tangent_offset(&self, _offset: f32) -> Vector2 {
        self.tangent_normalized(0.0) // tangent is constant for this type of alignment
    }
    
    fn length(&self) -> f32 {
        (self.upper_end - self.lower_end).length()
    }
}

pub struct TangentElevationSampler {
    lower_end: f32,
    upper_end: f32,
    top_down_length: f32,
}

impl TangentElevationSampler {
    pub fn new(lower_end: f32, upper_end: f32, top_down_length: f32) -> Self {
        if top_down_length <= 0.0 {
            panic!("TangentElevationSampler cannot have 0 length.");
        }
        Self {
            lower_end,
            upper_end,
            top_down_length,
        }
    }
}

impl ElevationSampler for TangentElevationSampler {
    fn sample_normalized(&self, top_down_t: f32) -> f32 {
        self.lower_end * (1.0 - top_down_t) + self.upper_end * top_down_t
    }
    
    fn grade_normalized(&self, _top_down_t: f32) -> f32 {
        (self.upper_end - self.lower_end) / self.top_down_length
    }
    
    fn sample_offset(&self, top_down_offset: f32) -> f32 {
        self.sample_normalized(top_down_offset / self.top_down_length)
    }
    
    fn grade_offset(&self, _top_down_offset: f32) -> f32 {
        self.grade_normalized(0.0)
    }
    
    fn length(&self) -> f32 {
        let rise = self.upper_end - self.lower_end;
        return (rise * rise + self.top_down_length * self.top_down_length).sqrt();
    }
    
    fn to_top_down_normalized(&self, actual_t: f32) -> f32 {
        actual_t
    }
    
    fn to_top_down_offset(&self, actual_offset: f32) -> f32 {
        self.top_down_length / self.length() * actual_offset
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    
    #[test]
    fn test_tangent_alignment_sample_offset() {
        let test_alignment_1 = TangentAlignmentSampler::new(Vector2::new(1.0, 1.0), Vector2::new(4.0, 5.0));
        assert_eq!(test_alignment_1.sample_offset(0.0), Vector2::new(1.0, 1.0),
                   "test_alignment_1 begins in the wrong location.");
        assert_eq!(test_alignment_1.sample_offset(5.0), Vector2::new(4.0, 5.0),
                   "test_alignment_1 ends in the wrong location.");
        assert_eq!(test_alignment_1.sample_offset(2.5), Vector2::new(2.5, 3.0),
                   "test_alignment_1 passes through the wrong location.");
        assert_eq!(test_alignment_1.sample_offset(-5.0), Vector2::new(-2.0, -3.0),
                   "test_alignment_1 negative offset ends in the wrong location.");
        assert_eq!(test_alignment_1.sample_offset(10.0), Vector2::new(7.0, 9.0),
                   "test_alignment_1 double offset ends in the wrong location.");
        let test_alignment_2 = TangentAlignmentSampler::new(Vector2::new(-1.0, -1.0), Vector2::new(-4.0, -5.0));
        assert_eq!(test_alignment_2.sample_offset(0.0), Vector2::new(-1.0, -1.0),
                   "test_alignment_2 begins in the wrong location.");
        assert_eq!(test_alignment_2.sample_offset(5.0), Vector2::new(-4.0, -5.0),
                   "test_alignment_2 ends in the wrong location.");
        assert_eq!(test_alignment_2.sample_offset(2.5), Vector2::new(-2.5, -3.0),
                   "test_alignment_2 passes through the wrong location.");
        assert_eq!(test_alignment_2.sample_offset(-5.0), Vector2::new(2.0, 3.0),
                   "test_alignment_2 negative offset ends in the wrong location.");
        assert_eq!(test_alignment_2.sample_offset(10.0), Vector2::new(-7.0, -9.0),
                   "test_alignment_2 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_alignment_tangent_offset() {
        let test_alignment_1 = TangentAlignmentSampler::new(Vector2::new(0.0, 0.0), Vector2::new(1.0, 0.0));
        let correct_tangent_1 = Vector2::new(1.0, 0.0);
        assert_eq!(test_alignment_1.tangent_offset(0.0), correct_tangent_1,
                   "test_alignment_1 begins with an incorrect tangent.");
        assert_eq!(test_alignment_1.tangent_offset(1.0), correct_tangent_1,
                   "test_alignment_1 ends with an incorrect tangent.");
        assert_eq!(test_alignment_1.tangent_offset(0.5), correct_tangent_1,
                   "test_alignment_1 has an incorrect tangent in the middle.");
        assert_eq!(test_alignment_1.tangent_offset(-1.0),correct_tangent_1,
                   "test_alignment_1 negative offset has an incorrect tangent.");
        assert_eq!(test_alignment_1.tangent_offset(2.0), correct_tangent_1,
                   "test_alignment_1 double offset has an incorrect tangent.");
        let test_alignment_2 = TangentAlignmentSampler::new(Vector2::new(0.0, 0.0), Vector2::new(0.0, -1.0));
        let correct_tangent_2 = Vector2::new(0.0, -1.0);
        assert_eq!(test_alignment_2.tangent_offset(0.0), correct_tangent_2,
                   "test_alignment_2 begins with an incorrect tangent.");
        assert_eq!(test_alignment_2.tangent_offset(1.0), correct_tangent_2,
                   "test_alignment_2 ends with an incorrect tangent.");
        assert_eq!(test_alignment_2.tangent_offset(0.5), correct_tangent_2,
                   "test_alignment_2 has an incorrect tangent in the middle.");
        assert_eq!(test_alignment_2.tangent_offset(-1.0), correct_tangent_2,
                   "test_alignment_2 negative offset has an incorrect tangent.");
        assert_eq!(test_alignment_2.tangent_offset(2.0), correct_tangent_2,
                   "test_alignment_2 double offset has an incorrect tangent.");
        let test_alignment_3 = TangentAlignmentSampler::new(Vector2::new(1.0, 1.0), Vector2::new(4.0, 5.0));
        let correct_tangent_3 = Vector2::new(3.0, 4.0).normalized();
        assert_eq!(test_alignment_3.tangent_offset(0.0), correct_tangent_3,
                   "test_alignment_3 begins with an incorrect tangent.");
        assert_eq!(test_alignment_3.tangent_offset(5.0), correct_tangent_3,
                   "test_alignment_3 ends with an incorrect tangent.");
        assert_eq!(test_alignment_3.tangent_offset(2.5), correct_tangent_3,
                   "test_alignment_3 has an incorrect tangent in the middle.");
        assert_eq!(test_alignment_3.tangent_offset(-5.0), correct_tangent_3,
                   "test_alignment_3 negative offset has an incorrect tangent.");
        assert_eq!(test_alignment_3.tangent_offset(10.0), correct_tangent_3,
                   "test_alignment_3 double offset has an incorrect tangent.");
    }
    
    #[test]
    fn test_tangent_alignment_sample_normalized() {
        let test_alignment_1 = TangentAlignmentSampler::new(Vector2::new(1.0, 1.0), Vector2::new(4.0, 5.0));
        assert_eq!(test_alignment_1.sample_normalized(0.0), Vector2::new(1.0, 1.0),
                   "test_alignment_1 begins in the wrong location.");
        assert_eq!(test_alignment_1.sample_normalized(1.0), Vector2::new(4.0, 5.0),
                   "test_alignment_1 ends in the wrong location.");
        assert_eq!(test_alignment_1.sample_normalized(0.5), Vector2::new(2.5, 3.0),
                   "test_alignment_1 passes through the wrong location.");
        assert_eq!(test_alignment_1.sample_normalized(-1.0), Vector2::new(-2.0, -3.0),
                   "test_alignment_1 negative offset ends in the wrong location.");
        assert_eq!(test_alignment_1.sample_normalized(2.0), Vector2::new(7.0, 9.0),
                   "test_alignment_1 double offset ends in the wrong location.");
        let test_alignment_2 = TangentAlignmentSampler::new(Vector2::new(-1.0, -1.0), Vector2::new(-4.0, -5.0));
        assert_eq!(test_alignment_2.sample_normalized(0.0), Vector2::new(-1.0, -1.0),
                   "test_alignment_2 begins in the wrong location.");
        assert_eq!(test_alignment_2.sample_normalized(1.0), Vector2::new(-4.0, -5.0),
                   "test_alignment_2 ends in the wrong location.");
        assert_eq!(test_alignment_2.sample_normalized(0.5), Vector2::new(-2.5, -3.0),
                   "test_alignment_2 passes through the wrong location.");
        assert_eq!(test_alignment_2.sample_normalized(-1.0), Vector2::new(2.0, 3.0),
                   "test_alignment_2 negative offset ends in the wrong location.");
        assert_eq!(test_alignment_2.sample_normalized(2.0), Vector2::new(-7.0, -9.0),
                   "test_alignment_2 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_alignment_tangent_normalized() {
        let test_alignment_1 = TangentAlignmentSampler::new(Vector2::new(0.0, 0.0), Vector2::new(1.0, 0.0));
        let correct_tangent_1 = Vector2::new(1.0, 0.0);
        assert_eq!(test_alignment_1.tangent_normalized(0.0),correct_tangent_1,
                   "test_alignment_1 begins with an incorrect tangent.");
        assert_eq!(test_alignment_1.tangent_normalized(1.0), correct_tangent_1,
                   "test_alignment_1 ends with an incorrect tangent.");
        assert_eq!(test_alignment_1.tangent_normalized(0.5), correct_tangent_1,
                   "test_alignment_1 has an incorrect tangent in the middle.");
        assert_eq!(test_alignment_1.tangent_normalized(-1.0), correct_tangent_1,
                   "test_alignment_1 negative offset has an incorrect tangent.");
        assert_eq!(test_alignment_1.tangent_normalized(2.0), correct_tangent_1,
                   "test_alignment_1 double offset has an incorrect tangent.");
        let test_alignment_2 = TangentAlignmentSampler::new(Vector2::new(0.0, 0.0), Vector2::new(0.0, -1.0));
        let correct_tangent_2 = Vector2::new(0.0, -1.0);
        assert_eq!(test_alignment_2.tangent_normalized(0.0), correct_tangent_2,
                   "test_alignment_2 begins with an incorrect tangent.");
        assert_eq!(test_alignment_2.tangent_normalized(1.0), correct_tangent_2,
                   "test_alignment_2 ends with an incorrect tangent.");
        assert_eq!(test_alignment_2.tangent_normalized(0.5), correct_tangent_2,
                   "test_alignment_2 has an incorrect tangent in the middle.");
        assert_eq!(test_alignment_2.tangent_normalized(-1.0), correct_tangent_2,
                   "test_alignment_2 negative offset has an incorrect tangent.");
        assert_eq!(test_alignment_2.tangent_normalized(2.0), correct_tangent_2,
                   "test_alignment_2 double offset has an incorrect tangent.");
        let test_alignment_3 = TangentAlignmentSampler::new(Vector2::new(1.0, 1.0), Vector2::new(4.0, 5.0));
        let correct_tangent_3 = Vector2::new(3.0, 4.0).normalized();
        assert_eq!(test_alignment_3.tangent_normalized(0.0), correct_tangent_3,
                   "test_alignment_3 begins with an incorrect tangent.");
        assert_eq!(test_alignment_3.tangent_normalized(1.0), correct_tangent_3,
                   "test_alignment_3 ends with an incorrect tangent.");
        assert_eq!(test_alignment_3.tangent_normalized(0.5), correct_tangent_3,
                   "test_alignment_3 has an incorrect tangent in the middle.");
        assert_eq!(test_alignment_3.tangent_normalized(-1.0), correct_tangent_3,
                   "test_alignment_3 negative offset has an incorrect tangent.");
        assert_eq!(test_alignment_3.tangent_normalized(2.0), correct_tangent_3,
                   "test_alignment_3 double offset has an incorrect tangent.");
    }
    
    #[test]
    fn test_tangent_alignment_length() {
        let test_alignment_1 = TangentAlignmentSampler::new(Vector2::new(1.0, 1.0), Vector2::new(4.0, 5.0));
        assert_eq!(test_alignment_1.length(), 5.0, "test_alignment_1 had an incorrect length.");
    }
    
    #[test]
    fn test_tangent_elevation_sample_offset() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 2.0, 5.0);
        assert_eq!(test_elevation_1.sample_offset(0.0), 1.0,
                   "test_elevation_1 begins in the wrong location.");
        assert_eq!(test_elevation_1.sample_offset(5.0), 2.0,
                   "test_elevation_1 ends in the wrong location.");
        assert_eq!(test_elevation_1.sample_offset(2.5), 1.5,
                   "test_elevation_1 passes through the wrong location.");
        assert_eq!(test_elevation_1.sample_offset(-5.0), 0.0,
                   "test_elevation_1 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_1.sample_offset(10.0), 3.0,
                   "test_elevation_1 double offset ends in the wrong location.");
        let test_elevation_2 = TangentElevationSampler::new(2.0, 2.0, 5.0);
        assert_eq!(test_elevation_2.sample_offset(0.0), 2.0,
                   "test_elevation_2 begins in the wrong location.");
        assert_eq!(test_elevation_2.sample_offset(5.0), 2.0,
                   "test_elevation_2 ends in the wrong location.");
        assert_eq!(test_elevation_2.sample_offset(2.5), 2.0,
                   "test_elevation_2 passes through the wrong location.");
        assert_eq!(test_elevation_2.sample_offset(-5.0), 2.0,
                   "test_elevation_2 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_2.sample_offset(10.0), 2.0,
                   "test_elevation_2 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_elevation_grade_offset() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 2.0, 5.0);
        assert_eq!(test_elevation_1.grade_offset(0.0), 0.2,
                   "test_elevation_1 begins in the wrong location.");
        assert_eq!(test_elevation_1.grade_offset(5.0), 0.2,
                   "test_elevation_1 ends in the wrong location.");
        assert_eq!(test_elevation_1.grade_offset(2.5), 0.2,
                   "test_elevation_1 passes through the wrong location.");
        assert_eq!(test_elevation_1.grade_offset(-5.0), 0.2,
                   "test_elevation_1 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_1.grade_offset(10.0), 0.2,
                   "test_elevation_1 double offset ends in the wrong location.");
        let test_elevation_2 = TangentElevationSampler::new(2.0, 2.0, 5.0);
        assert_eq!(test_elevation_2.grade_offset(0.0), 0.0,
                   "test_elevation_2 begins in the wrong location.");
        assert_eq!(test_elevation_2.grade_offset(5.0), 0.0,
                   "test_elevation_2 ends in the wrong location.");
        assert_eq!(test_elevation_2.grade_offset(2.5), 0.0,
                   "test_elevation_2 passes through the wrong location.");
        assert_eq!(test_elevation_2.grade_offset(-5.0), 0.0,
                   "test_elevation_2 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_2.grade_offset(10.0), 0.0,
                   "test_elevation_2 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_elevation_sample_normalized() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 2.0, 5.0);
        assert_eq!(test_elevation_1.sample_normalized(0.0), 1.0,
                   "test_elevation_1 begins in the wrong location.");
        assert_eq!(test_elevation_1.sample_normalized(1.0), 2.0,
                   "test_elevation_1 ends in the wrong location.");
        assert_eq!(test_elevation_1.sample_normalized(0.5), 1.5,
                   "test_elevation_1 passes through the wrong location.");
        assert_eq!(test_elevation_1.sample_normalized(-1.0), 0.0,
                   "test_elevation_1 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_1.sample_normalized(2.0), 3.0,
                   "test_elevation_1 double offset ends in the wrong location.");
        let test_elevation_2 = TangentElevationSampler::new(2.0, 2.0, 5.0);
        assert_eq!(test_elevation_2.sample_normalized(0.0), 2.0,
                   "test_elevation_2 begins in the wrong location.");
        assert_eq!(test_elevation_2.sample_normalized(1.0), 2.0,
                   "test_elevation_2 ends in the wrong location.");
        assert_eq!(test_elevation_2.sample_normalized(0.5), 2.0,
                   "test_elevation_2 passes through the wrong location.");
        assert_eq!(test_elevation_2.sample_normalized(-1.0), 2.0,
                   "test_elevation_2 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_2.sample_normalized(2.0), 2.0,
                   "test_elevation_2 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_elevation_grade_normalized() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 2.0, 5.0);
        assert_eq!(test_elevation_1.grade_normalized(0.0), 0.2,
                   "test_elevation_1 begins in the wrong location.");
        assert_eq!(test_elevation_1.grade_normalized(1.0), 0.2,
                   "test_elevation_1 ends in the wrong location.");
        assert_eq!(test_elevation_1.grade_normalized(0.5), 0.2,
                   "test_elevation_1 passes through the wrong location.");
        assert_eq!(test_elevation_1.grade_normalized(-1.0), 0.2,
                   "test_elevation_1 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_1.grade_normalized(2.0), 0.2,
                   "test_elevation_1 double offset ends in the wrong location.");
        let test_elevation_2 = TangentElevationSampler::new(2.0, 2.0, 5.0);
        assert_eq!(test_elevation_2.grade_normalized(0.0), 0.0,
                   "test_elevation_2 begins in the wrong location.");
        assert_eq!(test_elevation_2.grade_normalized(1.0), 0.0,
                   "test_elevation_2 ends in the wrong location.");
        assert_eq!(test_elevation_2.grade_normalized(0.5), 0.0,
                   "test_elevation_2 passes through the wrong location.");
        assert_eq!(test_elevation_2.grade_normalized(-1.0), 0.0,
                   "test_elevation_2 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_2.grade_normalized(2.0), 0.0,
                   "test_elevation_2 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_elevation_length() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 4.0, 4.0);
        assert_eq!(test_elevation_1.length(), 5.0, "test_elevation_1 had an incorrect length.");
    }
    
    #[test]
    fn test_tangent_elevation_to_top_down_normalized() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 4.0, 4.0);
        assert_eq!(test_elevation_1.to_top_down_normalized(0.0), 0.0,
                   "test_elevation_1 begins in the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_normalized(1.0), 1.0,
                   "test_elevation_1 ends in the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_normalized(0.5), 0.5,
                   "test_elevation_1 passes through the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_normalized(-1.0), -1.0,
                   "test_elevation_1 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_normalized(2.0), 2.0,
                   "test_elevation_1 double offset ends in the wrong location.");
    }
    
    #[test]
    fn test_tangent_elevation_to_top_down_offset() {
        let test_elevation_1 = TangentElevationSampler::new(1.0, 4.0, 4.0);
        assert_eq!(test_elevation_1.to_top_down_offset(0.0), 0.0,
                   "test_elevation_1 begins in the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_offset(5.0), 4.0,
                   "test_elevation_1 ends in the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_offset(2.5), 2.0,
                   "test_elevation_1 passes through the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_offset(-5.0), -4.0,
                   "test_elevation_1 negative offset ends in the wrong location.");
        assert_eq!(test_elevation_1.to_top_down_offset(10.0), 8.0,
                   "test_elevation_1 double offset ends in the wrong location.");
    }
}
